#ifndef TIN_AUTOCONF_H
#       include "autoconf.h"
#endif /* !TIN_AUTOCONF_H */

/* Define to the full name of this package. */
#define PACKAGE_NAME "libcanlock"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "libcanlock 3.0.0"
